from .basetool import BaseTool


class Marker(BaseTool):
    type = 'marker'
    handleCount = 1
    displayName = 'Target'

    def get_show_label(self):
        return self.get_option('show_label', 'true')

    def set_show_label(self, value):
        return self.set_option('show_label', value)



from .basetool import BaseTool


class Arrow(BaseTool):
    type = 'arrow'
    handleCount = 2
    displayName = 'Arrow'

    def get_color(self):
        return self.get_option('color', 'theme.toolArrowRedBackgroundColor')

    def set_color(self, value):
        return self.set_option('color', value)

    def get_borderColor(self):
        return self.get_option('borderColor', 'theme.toolArrowRedBorderColor')

    def set_borderColor(self, value):
        return self.set_option('borderColor', value)



def timeframe_align(input: int, timeframe: int) -> int:
    remainder = input % timeframe
    if remainder < timeframe / 2:
        # Round down if the remainder is less than half of 1920
        aligned_timestamp = input - remainder
    else:
        # Round up if the remainder is greater than or equal to half of 1920
        aligned_timestamp = input + (timeframe - remainder)

    return aligned_timestamp



class Checklist:
    def __init__(self, tpro):
        self.tpro = tpro

    def get_scores(self, turningpoint_id: str) -> str:
        data = self.tpro.rest.get(f"indicator/mtops/score/{turningpoint_id}")
        return data['enabled']

    def get_list(self):
        data = self.tpro.rest.get("indicator/mtops/list")
        return data



from .basetool import BaseTool


class TurningPoint(BaseTool):
    type = 'turningpoint'
    handleCount = 1
    displayName = 'Turning Point'

    def get_rays(self):
        return self.get_option('rays', '0')

    def set_rays(self, value):
        return self.set_option('rays', value)

    def get_fixedscale(self):
        return self.get_option('fixedscale', '0')

    def set_fixedscale(self, value):
        return self.set_option('fixedscale', value)

    def get_gannbox(self):
        return self.get_option('gannbox', '0')

    def set_gannbox(self, value):
        return self.set_option('gannbox', value)

    def get_doublearcs(self):
        return self.get_option('doublearcs', '0')

    def set_doublearcs(self, value):
        return self.set_option('doublearcs', value)

    def get_show_fib(self):
        return self.get_option('show_fib', '1')

    def set_show_fib(self, value):
        return self.set_option('show_fib', value)

    def get_grid(self):
        return self.get_option('grid', '0')

    def set_grid(self, value):
        return self.set_option('grid', value)

    def get_non_probability(self):
        return self.get_option('non_probability', '0')

    def set_non_probability(self, value):
        return self.set_option('non_probability', value)

    def get_extend_arcs(self):
        return self.get_option('extend_arcs', '0')

    def set_extend_arcs(self, value):
        return self.set_option('extend_arcs', value)

    def get_level(self):
        return self.get_option('level', '1')

    def set_level(self, value):
        return self.set_option('level', value)

    def get_number(self):
        return self.get_option('number', '0')

    def set_number(self, value):
        return self.set_option('number', value)



from .basetool import BaseTool


class Protractor(BaseTool):
    type = 'protractor'
    handleCount = 1
    displayName = 'Protractor'

    def get_type(self):
        return self.get_option('type', '2')

    def set_type(self, value):
        return self.set_option('type', value)

    def get_doublearcs(self):
        return self.get_option('doublearcs', '0')

    def set_doublearcs(self, value):
        return self.set_option('doublearcs', value)



from .basetool import BaseTool


class PitchFork(BaseTool):
    type = 'pitchfork'
    handleCount = 3
    displayName = 'Pitch Fork'

    def get_color(self):
        return self.get_option('color', 'theme.PitchForkLineColor')

    def set_color(self, value):
        return self.set_option('color', value)

    def get_width(self):
        return self.get_option('width', '1')

    def set_width(self, value):
        return self.set_option('width', value)

    def get_style(self):
        return self.get_option('style', 'solid')

    def set_style(self, value):
        return self.set_option('style', value)

    def get_levels(self):
        return self.get_option('levels', '[{"visible":false,"level":0.25},{"visible":false,"level":0.382},{"visible":true,"level":0.5},{"visible":false,"level":0.618},{"visible":false,"level":0.75},{"visible":true,"level":1},{"visible":false,"level":1.5},{"visible":false,"level":1.75},{"visible":false,"level":2}]')

    def set_levels(self, value):
        return self.set_option('levels', value)



from .basetool import BaseTool


class Image(BaseTool):
    type = 'image'
    handleCount = 2
    displayName = 'Image'

    def get_imageSrc(self):
        return self.get_option('imageSrc', '')

    def set_imageSrc(self, value):
        return self.set_option('imageSrc', value)

    def get_opacity(self):
        return self.get_option('opacity', '100')

    def set_opacity(self, value):
        return self.set_option('opacity', value)




def register_socketio_events(instance, socket) -> None:
    """
    Register all @on event handlers for a class instance
    with a Socket.IO client.
    """
    for attr_name in dir(instance):
        attr = getattr(instance, attr_name)
        if callable(attr) and hasattr(attr, "_socketio_event"):
            event_name, namespace = getattr(attr, "_socketio_event")
            socket.on(event_name, attr, namespace=namespace)



class ApiTokenError(Exception):
    """Raises when the API token is invalid"""
    pass


class AuthenticationTimeoutError(Exception):
    pass


class ToolHandleCountError(Exception):
    pass


class ToolExistsError(Exception):
    pass



import sys
import time
import socketio
import math

from .tools import Tools
from .markets import Markets
from .alerts import Alerts
from .restclient import RestClient
from .decorators.on import on
from .utils.socketio_utils import register_socketio_events
from .errors import ApiTokenError, AuthenticationTimeoutError
from .indicators.mtops.checklist import Checklist
from .timewindow import TimeWindow
from .chart import Chart

# Make sure we're not running inside an unsupported environment
if sys.version_info < (3, 9):
    raise RuntimeError(
        "The Tesseract Pro Python API requires Python 3.9 or higher.")


class TesseractPro:
    authenticated = False

    def __init__(self, api_token, server_url="https://piper.tesseractpro.io"):
        self.api_token = api_token
        self.server_url = server_url
        self._socket = socketio.Client()

        # Register TesseractPro events
        register_socketio_events(self, self._socket)

        self._socket.connect(
            self.server_url,
            namespaces=['/api'],
            auth={'token': self.api_token}
        )

        timeout = 10
        start_time = time.time()
        while not self.authenticated:
            timed_out = time.time() - start_time > timeout
            if timed_out or not self._socket.connected:
                if self._socket.connected:
                    raise AuthenticationTimeoutError(
                        "Authentication timed out")
                else:
                    raise ApiTokenError("Invalid token provided")

                self._socket.disconnect()
                exit(1)
            time.sleep(0.1)

        self.tool = Tools(self)
        self.market = Markets(self)
        self.rest = RestClient(self)
        self.alert = Alerts(self)

    def get_indicator(self, name: str):
        match name:
            case "mtops":
                return Checklist(self)

    def get_chart(
        self,
        space_id: str,
        timeframe: int,
        symbol: str,
        start_time: int,
        end_time: int,
    ) -> Chart:
        candles = math.ceil((start_time - end_time) / (timeframe * 60))
        time_window = TimeWindow(timeframe, start_time, candles)

        return Chart(
            tesseract_pro=self,
            symbol=symbol,
            space_id=space_id,
            timeframe=timeframe,
            time_window=time_window
        )

    def disconnect(self) -> None:
        self.authenticated = False
        self._socket.disconnect()

    def wait_for_events(self) -> None:
        self._socket.wait()

    @on("invalid_api_token")
    def on_token_error(self) -> None:
        self.disconnect()

    @on("authenticated")
    def on_authenticated(self, data) -> None:
        self.authenticated = True

    @on("disconnect")
    def on_disconnect(self) -> None:
        self.authenticated = False

    @on("error")
    def on_connect_error(self, data) -> None:
        self.authenticated = False

    def close(self) -> None:
        self.authenticated = False
        self._socket.disconnect()



from typing import Callable


def on(
    event_name,
    namespace="/api"
) -> Callable[[Callable[..., None]], Callable[..., None]]:
    """
    Decorator to mark a method as a Socket.IO event handler.
    """
    def decorator(func):
        setattr(func, "_socketio_event", (event_name, namespace))
        return func
    return decorator



import json
from ..handle import Handle


class BaseTool:
    handleCount = 1

    # Is the tool attached to a chart?
    chart = None

    def __init__(self, data=None):
        if data is None:
            data = {"id": None, "handles": []}
        print(f"BaseTool: {data}")
        self.data = data

    def get_id(self) -> str:
        if 'id' not in self.data:
            return None

        return self.data['id']

    def set_id(self, id):
        self.data['id'] = id

    def get_type(self) -> str:
        if hasattr(self, 'type'):
            return self.type

        return self.data['type']

    def get_option(self, key, default_value=""):
        if key not in self.data['options']:
            return default_value

        return json.loads(self.data['options'][key])

    def set_option(self, key, value):
        if 'options' not in self.data:
            self.data['options'] = {}

        self.data['options'][key] = json.dumps(value)

    def save_options(self):
        if self.chart is None:
            raise Exception("Tool is not attached to a chart")

        tpro = self.chart.tpro

        tpro.rest.post(f"tools/options/{self.get_id()}", {
            "options": self.data['options']
        })

        print(f"Saving options {self.data['options']}")

    def available_options(self) -> dict:
        return self.data['options'].keys()

    def set_handles(self, handles):
        if len(handles) == 0:
            return

        self.data['time'] = handles[0][0]
        self.data['price'] = handles[0][1]

        if len(handles) > 1:
            if 'handles' not in self.data:
                self.data['handles'] = []

            for handle in handles[1:]:
                self.data['handles'].append({
                    'time': handle[0],
                    'price': handle[1]
                })

    def get_handles(self) -> list:
        handles = []

        if 'time' in self.data and 'price' in self.data:
            handles.append(Handle(self.data['time'], self.data['price']))

        if 'handles' in self.data:
            for handle in self.data['handles']:
                handles.append(Handle(handle['time'], handle['price']))

        return handles



from .utils.event_emitter import EventEmitter
from .utils.socketio_utils import register_socketio_events


class Base (EventEmitter):
    def __init__(self, tesseract_pro):
        super().__init__()

        register_socketio_events(self, tesseract_pro._socket)

        self._tpro = tesseract_pro




from .turningpoint import TurningPoint
from .image import Image
from .magbox import MagBox
from .spiral import Spiral
from .marker import Marker
from .curve import Curve
from .line import Line
from .circle import Circle
from .angleline import AngleLine
from .vertical_ray import VerticalRay
from .horizontal_ray import HorizontalRay
from .arrow import Arrow
from .text import Text
from .protractor import Protractor
from .gann_grid import GannGrid
from .gann_square import GannSquare
from .gann_box import GannBox
from .gannfan import GannFan
from .fib_retracement import FibRetracement
from .pitchfork import PitchFork


from .basetool import BaseTool


class MagBox(BaseTool):
    type = 'magbox'
    handleCount = 2
    displayName = 'Rectangle'

    def get_color(self):
        return self.get_option('color', 'rgba(235, 227, 12, 0.1)')

    def set_color(self, value):
        return self.set_option('color', value)

    def get_borderColor(self):
        return self.get_option('borderColor', 'rgb(171,170,41)')

    def set_borderColor(self, value):
        return self.set_option('borderColor', value)

    def get_width(self):
        return self.get_option('width', '2')

    def set_width(self, value):
        return self.set_option('width', value)

    def get_style(self):
        return self.get_option('style', 'solid')

    def set_style(self, value):
        return self.set_option('style', value)




from .all_tools import *

from ..base import Base
from ..decorators.on import on
from ..utils import timeframe_align


class Tools (Base):
    @on("tool:add")
    def on_tools_added(self, data) -> None:
        self.emit('add', data)

    @on("tool:update")
    def on_tool_update(self, data) -> None:
        self.emit('update', data)

    @on("tool:remove")
    def on_tool_remove(self, data) -> None:
        self.emit('remove', data)

    def remove_tool(self, space_id, tool):
        self._tpro.rest.delete(f"tools/{space_id}/{tool.get_id()}")

    def add_tool(
        self,
        tool,
        timeframe,
        symbol,
        space_id
    ):
        handles = tool.get_handles()

        data = self._tpro.rest.put("tools", {
            "timeframe": timeframe,
            "market": symbol,
            "space_id": space_id,
            "time": timeframe_align(handles[0].get_time(), timeframe),
            "price": handles[0].get_price(),
            "type": tool.get_type(),
            "options": tool.data.get('options'),
            "handles": [{"price": item.get_price(), "time": timeframe_align(item.get_time(), timeframe)} for item in handles[1:]]
        })

        return data['id']

    def get_tools(
        self,
        from_unix: int,
        to_unix: int,
        timeframe: int,
        market: str,
        space_id: str
    ):
        """
        Return tools for the given parameters
        """
        data = self._tpro.rest.post("tools", {
            "to": to_unix,
            "from": from_unix,
            "timeframe": timeframe * 60,
            "market": market,
            "space_id": space_id
        })

        return {item['id']: item for item in data}



from .basetool import BaseTool


class Text(BaseTool):
    type = 'text'
    handleCount = 1
    displayName = 'Text'

    def get_adaptive(self):
        return self.get_option('adaptive', 'false')

    def set_adaptive(self, value):
        return self.set_option('adaptive', value)

    def get_color(self):
        return self.get_option('color', 'theme.textColor')

    def set_color(self, value):
        return self.set_option('color', value)

    def get_fontsize(self):
        return self.get_option('fontsize', '24')

    def set_fontsize(self, value):
        return self.set_option('fontsize', value)

    def get_text(self):
        return self.get_option('text', 'Type in Text Properties')

    def set_text(self, value):
        return self.set_option('text', value)



from setuptools import setup, find_packages

setup(
    name="tesseractpro",
    version="0.1.0",
    description="The Tesseract Pro Python API",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Emiel van Goor",
    author_email="info@tesseractpro.io",
    url="https://github.com/tesseractpro/python-api",
    packages=find_packages(include=["tesseractpro", "tesseractpro.*"]),
    install_requires=[
        "python-socketio[client]>=5.0.0",
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
    license="MIT",
    keywords="TesseractPro API",
)



from .basetool import BaseTool


class FibRetracement(BaseTool):
    type = 'fib-retracement'
    handleCount = 2
    displayName = 'Fibonacci Retracement'

    def get_extendLeft(self):
        return self.get_option('extendLeft', '0')

    def set_extendLeft(self, value):
        return self.set_option('extendLeft', value)

    def get_extendRight(self):
        return self.get_option('extendRight', '0')

    def set_extendRight(self, value):
        return self.set_option('extendRight', value)

    def get_trendline(self):
        return self.get_option('trendline', '0')

    def set_trendline(self, value):
        return self.set_option('trendline', value)

    def get_levels(self):
        return self.get_option('levels', '[{ "visible": true, "level": 0 }, { "visible": true, "level": 0.236 }, { "visible": true, "level": 0.382 }, { "visible": true, "level": 0.5 }, { "visible": true, "level": 0.618 }, { "visible": true, "level": 0.786 }, { "visible": true, "level": 1 }, { "visible": true, "level": 1.618 }, { "visible": true, "level": 2.618 } ]')

    def set_levels(self, value):
        return self.set_option('levels', value)

    def get_color(self):
        return self.get_option('color', 'theme.FibRetracementLineColor')

    def set_color(self, value):
        return self.set_option('color', value)

    def get_width(self):
        return self.get_option('width', '1')

    def set_width(self, value):
        return self.set_option('width', value)

    def get_style(self):
        return self.get_option('style', 'solid')

    def set_style(self, value):
        return self.set_option('style', value)



from .basetool import BaseTool


class GannGrid(BaseTool):
    type = 'gann-grid'
    handleCount = 2
    displayName = 'Gann Grid'

    def get_show_grid(self):
        return self.get_option('show_grid', 'true')

    def set_show_grid(self, value):
        return self.set_option('show_grid', value)



from .basetool import BaseTool


class GannSquare(BaseTool):
    type = 'gann-square'
    handleCount = 2
    displayName = 'Gann Square'

    def get_color(self):
        return self.get_option('color', 'theme.gannSquareLineColor')

    def set_color(self, value):
        return self.set_option('color', value)

    def get_gannbox(self):
        return self.get_option('gannbox', '0')

    def set_gannbox(self, value):
        return self.set_option('gannbox', value)

    def get_doublearcs(self):
        return self.get_option('doublearcs', '0')

    def set_doublearcs(self, value):
        return self.set_option('doublearcs', value)



from .basetool import BaseTool


class Spiral(BaseTool):
    type = 'spiral'
    handleCount = 2
    displayName = 'Spiral'

    def get_show_rectangles(self):
        return self.get_option('show_rectangles', 'false')

    def set_show_rectangles(self, value):
        return self.set_option('show_rectangles', value)

    def get_width(self):
        return self.get_option('width', '1')

    def set_width(self, value):
        return self.set_option('width', value)

    def get_rectangle_color(self):
        return self.get_option('rectangle_color', 'theme.SpiralRectangleColor')

    def set_rectangle_color(self, value):
        return self.set_option('rectangle_color', value)

    def get_color(self):
        return self.get_option('color', 'color.primary')

    def set_color(self, value):
        return self.set_option('color', value)



class Handle:
    def __init__(self, time, price):
        self.time = time
        self.price = price

    def get_time(self):
        return self.time

    def get_price(self):
        return self.price



class EventEmitter:
    def __init__(self):
        self._events = {}

    def on(self, event_name, callback) -> None:
        """
        Register a callback for a specific event.
        """
        if event_name not in self._events:
            self._events[event_name] = []
        self._events[event_name].append(callback)

    def emit(self, event_name, *args, **kwargs) -> None:
        """
        Invoke all callbacks associated with an event.
        """
        if event_name in self._events:
            for callback in self._events[event_name]:
                callback(*args, **kwargs)



from .basetool import BaseTool


class GannBox(BaseTool):
    type = 'gann-box'
    handleCount = 2
    displayName = 'Gann Box'

    def get_color(self):
        return self.get_option('color', 'theme.gannBoxLineColor')

    def set_color(self, value):
        return self.set_option('color', value)

    def get_width(self):
        return self.get_option('width', '1')

    def set_width(self, value):
        return self.set_option('width', value)

    def get_style(self):
        return self.get_option('style', 'solid')

    def set_style(self, value):
        return self.set_option('style', value)

    def get_labels(self):
        return self.get_option('labels', 'true')

    def set_labels(self, value):
        return self.set_option('labels', value)

    def get_levels_time(self):
        return self.get_option('levels_time', '[{ "visible": true, "angle": false, "level": 0.125 }, { "visible": true, "angle": false, "level": 0.25 }, { "visible": true, "angle": false, "level": 0.382 }, { "visible": true, "angle": true, "level": 0.5 }, { "visible": true, "angle": false, "level": 0.618 }, { "visible": true, "angle": false, "level": 0.786 }, { "visible": true, "angle": false, "level": 0.886 }]')

    def set_levels_time(self, value):
        return self.set_option('levels_time', value)

    def get_levels_price(self):
        return self.get_option('levels_price', '[{ "visible": true, "angle": false, "level": 0.125 }, { "visible": true, "angle": false, "level": 0.25 }, { "visible": true, "angle": false, "level": 0.382 }, { "visible": true, "angle": true, "level": 0.5 }, { "visible": true, "angle": false, "level": 0.618 }, { "visible": true, "angle": false, "level": 0.786 }, { "visible": true, "angle": false, "level": 0.886 }]')

    def set_levels_price(self, value):
        return self.set_option('levels_price', value)



from .basetool import BaseTool


class GannFan(BaseTool):
    type = 'gannfan'
    handleCount = 2
    displayName = 'Gann Fan'

    def get_color(self):
        return self.get_option('color', 'color.primary')

    def set_color(self, value):
        return self.set_option('color', value)

    def get_width(self):
        return self.get_option('width', '1')

    def set_width(self, value):
        return self.set_option('width', value)

    def get_style(self):
        return self.get_option('style', 'solid')

    def set_style(self, value):
        return self.set_option('style', value)

    def get_levels(self):
        return self.get_option('levels', '[{"visible": true,"level":8},{"visible": true,"level":4},{"visible": true,"level":3},{"visible": true,"level":2},{"visible": true,"level":1}]')

    def set_levels(self, value):
        return self.set_option('levels', value)
